%
% qratser.m  Date: 5/20/2010
% This Matlab program computes E[(x'Ax)^p/(x'Bx)^q], where p is a
% nonnegative real number, q is a nonnegative real number, and
% z ~ N(0_n,I_n).  It makes use of an infinite series approach.
% Input:
% A: matrix of the quadratic form in the numerator
% B: matrix of the quadratic form in the denominator
% p: power of the quadratic form in the numerator
% q: power of the quadratic form in the denominator
% E: relative error (default is 1e-10)
% Output:
% y: E[(z'Az)^p/(z'Bz)^q]
% j: Number of iterations
%
function [y,j] = qratser(A,B,p,q,E);
j = 0;
if nargin<5
   E = 1e-10;
end
if nargin<4
   q = p;
end
n = size(A,1);
if n==1
   if p+0.5>q
      y = (2*A)^p/(2*B)^q*gamma(0.5+p-q)/sqrt(pi);
   else
      y = NaN;
   end
   return
end
pr = ceil(p)-p;
M = 1000;  % Maximum number of terms
if q==0
   if p==0
      y = 1;
   else
      a = 2*eig(A);
      ra = rank(A);
      if ra<n
         [~,i1] = sort(abs(a));
         a(i1(1:n-ra)) = 0;
      end
      if pr==0
         q0 = zeros(n,1);
         d = 1;
         for i=1:p
             q0 = a.*(d+q0);
             d = sum(q0)/(2*i);
         end
         y = gamma(p+1)*d;
      elseif any(a<0)
         y = NaN;
         fprintf('A matrix must be positive semi-definite when p is not an integer!\n')
      else
         maxea = max(a);
         a = 1-a/maxea;
         a(a==0) = [];
         n2 = n/2;
         c = maxea^p*gamma(n2+p)/gamma(n2);
         q0 = 0;
         q1 = c;
         y = q1;
         d = 1;
         q2 = zeros(size(a));
         while (abs(q1)>q0||abs(q1/y)>E)&&j<M
            q0 = abs(q1);
            c = c*(-p+j)/(n2+j);
            j = j+1;
            q2 = a.*(d+q2);
            d = sum(q2)/(2*j);
            q1 = c*d;
            y = y+q1;
         end
         if j==M
            fprintf(' Maximum number of terms reached\n')
         end
      end
   end
   return
end
[P,D] = eig(B);
d = diag(D);
rb = rank(B);
if rb<n
   [~,i1] = sort(abs(d));
   d(i1(1:n-rb)) = 0;
end
sn = 1;
if all(d<=0)
   if rem(q,1)==0
      d = abs(d);
      sn = 1-2*rem(q,2);   % sn=-1 if q is odd, and +1 if q is even
   else
      y = NaN;
      fprintf('B matrix must be positive semi-definite when q is not an integer!\n')
      return
   end
end
[d,ii] = sort(d,1,'descend');
if d(1)*d(end)<0
   y = NaN;
   fprintf('B matrix must be semi-definite!\n')
   return
end
maxb = d(1);
sn = sn/maxb^q;
d = d/maxb;
da = eig(A);
ra = rank(A);
if ra<n
   [~,i1] = sort(abs(da));
   da(i1(1:n-ra)) = 0;
end
if pr>0&&any(da<0)
   y = NaN;
   fprintf('A matrix must be positive semi-definite when p is not an integer!\n')
   return
elseif all(da<=0)
   A = -A;
   da = abs(da);
   sn = sn*(1-2*rem(p,2));
end
maxa = max(da);
sn = sn*maxa^p;
A = A./maxa;
P = P(:,ii);
A = P'*A*P;
m = rank(B);
d(m+1:end) = 0;
%
%  Check for existence of moments
%
if m/2+p<=q
   y = NaN;
   return
end
if m<n   % Stronger condition for existence of moments when B is psd.
   A22 = A(m+1:end,m+1:end);
   A12 = A(1:m,m+1:end);
   if any(abs(A22(:))>eps)
      if m/2<=q
         y = NaN;
         return
      end
   elseif any(abs(A12(:))>eps)
      if (m+p)/2<=q
         y = NaN;
         return
      end
   else
%
%  reduce the size of the problem if A_{12}=0_{mx(n-m)} and A_{22}=0_{(n-m)x(n-m)}
%
      A = A(1:m,1:m);
      d = d(1:m);
      n = m;
   end
end
n2 = n/2;
eb = 1-d;
Bh = diag(eb);
D = eb*ones(1,n);
if pr==0
   G = zeros(n,n,p+1);
   d = ones(p+1,1);
   n2p = n2+p;
   c = 2^(p-q)*sn*exp(gammaln(p+1)+gammaln(n2p-q)-gammaln(n2p));
   for i=1:p
       G1 = A*d(i)+A*G(:,:,i);
       d(i+1) = trace(G1)/(2*i);
       G(:,:,i+1) = G1;
   end
   q1 = c*d(end);
   y = q1;
   q0 = 0;
   while (abs(q1)>q0||abs(q1/y)>E)&&j<M
      q0 = abs(q1);
      c = c*(q+j)/(n2p+j);
      j = j+1;
      G1 = d(1)*Bh+D.*G(:,:,1);
      d(1) = trace(G1)/(2*j);
      G(:,:,1) = G1;
      for i=1:p
          G1 = A*d(i)+Bh*d(i+1)+A*G1+D.*G(:,:,i+1);
          d(i+1) = trace(G1)/(2*(i+j));
          G(:,:,i+1) = G1;
      end
      q1 = c*d(end);
      y = y+q1;
   end
else
   Ah = eye(n)-A;
   q1 = 1;
   y = 1;
   q0 = 0;
% 
%  For a given j=0,1,...,
%  c1(i) = (-p)_{i-1}*(q)_{j-i+1}/(n/2)_j, i=1,...,j+1
%  The rescaling of c1 will cost additional computation 
%  time, but it will reduce the overflow problem.
%
   c1 = ones(M+1,1);
   d = ones(M+1,1);
   G = zeros(n);
   while (abs(q1)>q0||abs(q1/y)>E)&&j<M
      q0 = abs(q1);
      j = j+1;
      c1(j+1) = c1(j)*(-p+j-1)/(n2+j-1);
      c1(1:j) = c1(1:j).*[q+j-1:-1:q]'/(n2+j-1);
      G = [eb*ones(1,n*j).*G zeros(n)]+[zeros(n) Ah*G];
      G(:,1:n) = G(:,1:n)+d(1)*Bh;
      for i=2:j
          G(:,(i-1)*n+1:i*n) = G(:,(i-1)*n+1:i*n)+d(i)*Bh+d(i-1)*Ah;
      end
      G(:,j*n+1:(j+1)*n) = G(:,j*n+1:(j+1)*n)+d(j)*Ah;
      for i=1:j+1
          d(i) = trace(G(:,(i-1)*n+1:i*n))/(2*j);
      end
      q1 = sum(c1(1:j+1).*d(1:j+1));
      y = y+q1;
   end
   y = y*2^(p-q)*sn*exp(gammaln(n2+p-q)-gammaln(n2));
end
if j==M
   fprintf(' Maximum number of terms reached\n')
end
